from abc import ABCMeta, abstractmethod
from sugarbowl import Function
from .emitter import Emitter
from .error import NotFound


# TODO: remove protected interface?
class find(Function, metaclass=ABCMeta):
    """Find framework base class.

    Parameters
    ----------
    values: iterable
        Values to process.
    mappers: list
        List of mappers.
    reducers: list
        List of reducers.
    emitter: type
        Emitter class.
    getfirst: bool
        Return first if True.
    fallback: mixed
        Fallback if fails.

    Returns
    -------
    mixed
        Values after map_reduce.
    """

    # Public

    # TODO: remove default?
    default_emitter = Emitter
    default_getfirst_exception = NotFound

    # TODO: remove getfirst_exception from the interface?
    def __init__(self, *,
                 filters=None, constraints=None,
                 mappers=None, reducers=None, emitter=None,
                 getfirst=False, getfirst_exception=None, fallback=None):
        if filters is None:
            filters = []
        if constraints is None:
            constraints = []
        if mappers is None:
            mappers = []
        if reducers is None:
            reducers = []
        if emitter is None:
            emitter = self.default_emitter
        if getfirst_exception is None:
            getfirst_exception = self.default_getfirst_exception
        self.__filters = filters
        self.__constraints = constraints
        self.__mappers = mappers
        self.__reducers = reducers
        self.__emitter = emitter
        self.__getfirst = getfirst
        self.__getfirst_exception = getfirst_exception
        self.__fallback = fallback
        self.__init_constraints()

    def __call__(self):
        mapped_values = self._map(self._values)
        reduced_values = self._reduce(mapped_values)
        return reduced_values

    # Protected

    def _map(self, values):
        for emitter in values:
            if not isinstance(emitter, self._emitter):
                emitter = self._emitter(emitter)
            for mapper in self._mappers:
                mapper(emitter)
                if emitter.skipped:
                    break
            if emitter.skipped:
                if emitter.stopped:
                    break
                continue  # pragma: no cover (coverage bug)
            if emitter.emitted:
                yield from emitter.emitted
            else:
                yield emitter.value()
            if emitter.stopped:
                break

    def _reduce(self, values):
        try:
            result = values
            if self._getfirst:
                try:
                    result = next(result)
                except StopIteration:
                    raise self._getfirst_exception()
            for reducer in self._reducers:
                result = reducer(result)
            return result
        except Exception as exception:
            if isinstance(self._fallback, Exception):
                raise self._fallback
            elif callable(self._fallback):
                return self._fallback(exception)
            elif self._fallback is not None:
                return self._fallback
            else:
                raise

    @property
    @abstractmethod
    def _values(self):
        pass  # pragma: no cover

    @property
    def _filters(self):
        return self.__filters

    @property
    def _constraints(self):
        return self.__constraints

    @property
    def _mappers(self):
        mappers = []
        for constraint in self._constraints:
            if constraint:
                mappers.append(constraint)
        mappers += self.__mappers
        return mappers

    @property
    def _reducers(self):
        return self.__reducers

    @property
    def _emitter(self):
        return self.__emitter

    @property
    def _getfirst(self):
        return self.__getfirst

    @property
    def _getfirst_exception(self):
        return self.__getfirst_exception

    @property
    def _fallback(self):
        return self.__fallback

    # Private

    def __init_constraints(self):
        for element in self._filters:
            for name, value in element.items():
                for constraint in self._constraints:
                    constraint.extend(name, value)
